#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include "ygString.h"
#include "ygUtil.h"
#include "xptTokenStream.h"

using namespace std;


xptTokenStream::xptTokenStream(const ygString& filename,FILE *outfp)
	{
	char *searchPath = ".";
	ygString fullPath;
	if (getenv("YG_PATH"))
		searchPath = getenv("YG_PATH");
	if (ygFindFile(filename,searchPath,fullPath))
		{
		fp_ = fopen(fullPath.c_str(),"r");
		if (!fp_)
			{
			cerr << "ERROR: xptTokenStream could not open file '"
				<< filename << "'\n";
			perror(fullPath.c_str());
			}
		}
	else
		{
		cerr << "ERROR: xptTokenStream could not find file '"
			<< filename << "'\n";
		fp_ = NULL;
		}
	filename_ = filename;
	linebufptr_ = NULL;
	lineNumber_ = 0;
	outfp_ = outfp;
	}


xptTokenStream::~xptTokenStream(void)
	{
	if (fp_)
		fclose(fp_);
	}


void xptTokenStream::pushback(const ygString& str)
	{
	ygString * s = new ygString(str);
	tokenStack_.push(s);
	}


bool xptTokenStream::next(ygString & str,bool printSpace)
	{
	if (!tokenStack_.empty())
		{
		ygString *tmp = tokenStack_.top();
		tokenStack_.pop();
		str = *tmp;
		delete tmp;
		return true;
		}
	else
		{
		return readNextToken(str,printSpace);
		}
	}


bool xptTokenStream::readNextToken(ygString& token,bool printSpace)
	{
	int c;
	token.clear();
	/* Find the next non-whitespace character */
	c = nextChar(printSpace);
	while ((c != -1) && ((c == ' ') || (c == '\t')))
		{
		if (printSpace)
			fprintf(outfp_,"%c",c);
		c = nextChar(printSpace);
		}
	/* If it's a brace, paren, comma, or equals, just return that */
	if (strchr("(){}=,",c))
		token += (char)c;
	/* If it's a quote, return everything up to the close quote */
	else if (c == '"')
		{
		while (1)
			{
			c = nextChar(printSpace);
			if (c == -1)
				{
				cerr << "ERROR (xptTokenStream): unclosed "
				    "quoted string in file " << filename_ << endl;
				return true;
				}
			else if (c == '"')
				break;
			token += (char)c;
			}
		}
	/* For anything else, return up to the next separator */
	else if (c > -1)
		{
		while ((c > -1) && (!strchr("(){}=, \t\n",(char)c)))
			{
			token += (char)c;
			c = nextChar(printSpace);
			}
		if (c > -1)
			pushbackChar(c);
		}
	else
		return false;
	return true;
	}


int xptTokenStream::nextChar(bool printSpace)
	{
	if (!charStack_.empty())
		{
		int c = charStack_.top();
		charStack_.pop();
		return c;
		}
	else
		{
		int c;
		if ((!linebufptr_) || (!*linebufptr_))
			{
			if (!getNextLine(printSpace))
				return -1;
			linebufptr_ = linebuf_;
			}
		c = *linebufptr_;
		linebufptr_++;
		return c;
		}
	}


void xptTokenStream::pushbackChar(int c)
	{
	charStack_.push(c);
	}


bool xptTokenStream::getNextLine(bool printSpace)
	{
	if (!fp_)
		return false;
	linebuf_[sizeof(linebuf_)-1] = '\0';
	while (fgets(linebuf_,sizeof(linebuf_)-1,fp_))
		{
		char *p = linebuf_;
		if (prevComment_.length() > 0)
			fprintf(outfp_,"//%s\n", prevComment_.c_str());
		else if ((lineNumber_ > 0) && (printSpace))
			fprintf(outfp_,"\n");
		lineNumber_++;
		prevComment_.clear();
		while ((*p) && (*p != '#') && (*p != '\n'))
			p++;
		if (*p == '#')
			{
			char *q = p+1;
			while ((*q) && (*q!='\n'))
				{
				prevComment_ += *q;
				q++;
				}
			}
		*p = '\0';
		if (strspn(linebuf_," \t") < strlen(linebuf_))
			return true;
		}
	fclose(fp_);
	fp_ = NULL;
	return false;
	}
